# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

import aiohttp
import pytest
from testing_support.external_fixtures import create_incoming_headers
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_cross_process_headers import (
    validate_cross_process_headers,
)
from testing_support.validators.validate_external_node_params import (
    validate_external_node_params,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.external_trace import ExternalTrace
from newrelic.api.transaction import current_transaction

version_info = tuple(int(_) for _ in aiohttp.__version__.split(".")[:2])

if version_info < (2, 0):
    _expected_error_class = aiohttp.errors.HttpProcessingError
else:
    _expected_error_class = aiohttp.client_exceptions.ClientResponseError


async def fetch(url, headers=None, raise_for_status=False, connector=None):

    kwargs = {}
    if version_info >= (2, 0):
        kwargs = {"raise_for_status": raise_for_status}

    session = aiohttp.ClientSession(connector=connector, **kwargs)
    request = session._request("GET", url, headers=headers)
    headers = {}

    try:
        response = await request
        if raise_for_status and version_info < (2, 0):
            response.raise_for_status()
    except _expected_error_class:
        return headers

    response_text = await response.text()
    for header in response_text.split("\n"):
        if not header:
            continue
        try:
            h, v = header.split(":", 1)
        except ValueError:
            continue
        headers[h.strip()] = v.strip()
    f = session.close()
    await asyncio.ensure_future(f)
    return headers


@pytest.mark.parametrize("cat_enabled", (True, False))
@pytest.mark.parametrize("distributed_tracing", (True, False))
@pytest.mark.parametrize("span_events", (True, False))
def test_outbound_cross_process_headers(event_loop, cat_enabled, distributed_tracing, span_events, mock_header_server):
    @background_task(name="test_outbound_cross_process_headers")
    async def _test():
        headers = await fetch("http://127.0.0.1:%d" % mock_header_server.port)

        transaction = current_transaction()
        transaction._test_request_headers = headers

        if distributed_tracing:
            assert "newrelic" in headers
        elif cat_enabled:
            assert ExternalTrace.cat_id_key in headers
            assert ExternalTrace.cat_transaction_key in headers
        else:
            assert "newrelic" not in headers
            assert ExternalTrace.cat_id_key not in headers
            assert ExternalTrace.cat_transaction_key not in headers

        def _validate():
            pass

        if cat_enabled or distributed_tracing:
            _validate = validate_cross_process_headers(_validate)

        _validate()

    @override_application_settings(
        {
            "cross_application_tracer.enabled": cat_enabled,
            "distributed_tracing.enabled": distributed_tracing,
            "span_events.enabled": span_events,
        }
    )
    def test():
        event_loop.run_until_complete(_test())

    test()


_nr_key = ExternalTrace.cat_id_key
_customer_headers_tests = [
    {"Test-Header": "Test Data 1"},
    {_nr_key.title(): "Test Data 2"},
]


@pytest.mark.parametrize("customer_headers", _customer_headers_tests)
def test_outbound_cross_process_headers_custom_headers(event_loop, customer_headers, mock_header_server):

    headers = event_loop.run_until_complete(
        background_task()(fetch)("http://127.0.0.1:%d" % mock_header_server.port, customer_headers.copy())
    )

    # always honor customer headers
    for expected_header, expected_value in customer_headers.items():
        assert headers.get(expected_header) == expected_value


def test_outbound_cross_process_headers_no_txn(event_loop, mock_header_server):
    headers = event_loop.run_until_complete(fetch("http://127.0.0.1:%d" % mock_header_server.port))

    assert not headers.get(ExternalTrace.cat_id_key)
    assert not headers.get(ExternalTrace.cat_transaction_key)


def test_outbound_cross_process_headers_exception(event_loop, mock_header_server):
    @background_task(name="test_outbound_cross_process_headers_exception")
    async def test():
        # corrupt the transaction object to force an error
        transaction = current_transaction()
        guid = transaction.guid
        delattr(transaction, "guid")

        try:
            headers = await fetch("http://127.0.0.1:%d" % mock_header_server.port)

            assert not headers.get(ExternalTrace.cat_id_key)
            assert not headers.get(ExternalTrace.cat_transaction_key)
        finally:
            transaction.guid = guid

    event_loop.run_until_complete(test())


class PoorResolvingConnector(aiohttp.TCPConnector):
    async def _resolve_host(self, host, port, *args, **kwargs):
        res = [{"hostname": host, "host": host, "port": 1234, "family": self._family, "proto": 0, "flags": 0}]
        hosts = await super(PoorResolvingConnector, self)._resolve_host(host, port, *args, **kwargs)
        for hinfo in hosts:
            res.append(hinfo)
        return res


@pytest.mark.parametrize("cat_enabled", [True, False])
@pytest.mark.parametrize("response_code", [200, 404])
@pytest.mark.parametrize("raise_for_status", [True, False])
@pytest.mark.parametrize("connector_class", [None, PoorResolvingConnector])  # None will use default
def test_process_incoming_headers(
    event_loop, cat_enabled, response_code, raise_for_status, connector_class, mock_external_http_server
):

    # It was discovered via packnsend that the `throw` method of the `_request`
    # coroutine is used in the case of poorly resolved hosts. An older version
    # of the instrumentation ended the ExternalTrace anytime `throw` was called
    # which meant that incoming CAT headers were never processed. The
    # `PoorResolvingConnector` connector in this test ensures that `throw` is
    # always called and thus makes sure the trace is not ended before
    # StopIteration is called.
    server, response_values = mock_external_http_server
    address = "http://127.0.0.1:%d" % server.port
    port = server.port

    _test_cross_process_response_scoped_metrics = [
        ("ExternalTransaction/127.0.0.1:%d/1#2/test" % port, 1 if cat_enabled else None)
    ]

    _test_cross_process_response_rollup_metrics = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/127.0.0.1:%d/all" % port, 1),
        ("ExternalApp/127.0.0.1:%d/1#2/all" % port, 1 if cat_enabled else None),
        ("ExternalTransaction/127.0.0.1:%d/1#2/test" % port, 1 if cat_enabled else None),
    ]

    _test_cross_process_response_external_node_params = [
        ("cross_process_id", "1#2"),
        ("external_txn_name", "test"),
        ("transaction_guid", "0123456789012345"),
    ]

    _test_cross_process_response_external_node_forgone_params = [
        k for k, v in _test_cross_process_response_external_node_params
    ]

    connector = connector_class() if connector_class else None

    @background_task(name="test_process_incoming_headers")
    async def _test():
        transaction = current_transaction()
        headers = create_incoming_headers(transaction)

        response_values.append((headers, response_code))

        await fetch(address, raise_for_status=raise_for_status, connector=connector)

    @override_application_settings(
        {"cross_application_tracer.enabled": cat_enabled, "distributed_tracing.enabled": False}
    )
    @validate_transaction_metrics(
        "test_process_incoming_headers",
        scoped_metrics=_test_cross_process_response_scoped_metrics,
        rollup_metrics=_test_cross_process_response_rollup_metrics,
        background_task=True,
    )
    @validate_external_node_params(
        params=(_test_cross_process_response_external_node_params if cat_enabled else []),
        forgone_params=([] if cat_enabled else _test_cross_process_response_external_node_forgone_params),
    )
    def test():
        event_loop.run_until_complete(_test())

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import sys

from aiohttp import ClientSession, WSMsgType, web

from newrelic.api.function_trace import function_trace


async def index(request):
    await asyncio.sleep(0)
    resp = web.Response(text="Hello Aiohttp!")
    resp.set_cookie("ExampleCookie", "ExampleValue")
    return resp


async def hang(request):
    while True:
        await asyncio.sleep(0)


async def error(request):
    raise ValueError("Value Error")


async def non_500_error(request):
    raise web.HTTPGone()


async def raise_403(request):
    raise web.HTTPForbidden()


async def raise_404(request):
    raise web.HTTPNotFound()


@function_trace()
async def wait():
    await asyncio.sleep(0.1)


async def run_task(loop):
    await wait()
    loop.stop()


async def background(request):
    try:
        loop = request.loop
    except AttributeError:
        loop = request.task._loop

    asyncio.set_event_loop(loop)
    asyncio.tasks.ensure_future(run_task(loop))
    return web.Response(text="Background Task Scheduled")


class HelloWorldView(web.View):
    async def _respond(self):
        await asyncio.sleep(0)
        resp = web.Response(text="Hello Aiohttp!")
        resp.set_cookie("ExampleCookie", "ExampleValue")
        return resp

    get = _respond
    post = _respond
    put = _respond
    patch = _respond
    delete = _respond


class KnownException(Exception):
    pass


class KnownErrorView(web.View):
    async def _respond(self):
        try:
            await asyncio.sleep(0)
        except KnownException:
            pass
        finally:
            return web.Response(text="Hello Aiohttp!")

    get = _respond
    post = _respond
    put = _respond
    patch = _respond
    delete = _respond


async def websocket_handler(request):

    ws = web.WebSocketResponse()
    await ws.prepare(request)

    # receive messages for all eternity!
    # (or until the client closes the socket)
    while not ws.closed:
        msg = await ws.receive()
        if msg.type == WSMsgType.TEXT:
            result = ws.send_str("/" + msg.data)
            if hasattr(result, "__await__"):
                await result

    return ws


async def fetch(method, url, loop):
    session = ClientSession(loop=loop)

    if hasattr(session, "__aenter__"):
        await session.__aenter__()
    else:
        session.__enter__()

    try:
        _method = getattr(session, method)
        try:
            response = await asyncio.wait_for(_method(url), timeout=None, loop=loop)
        except TypeError:
            response = await asyncio.wait_for(_method(url), timeout=None)
        text = await response.text()

    finally:
        if hasattr(session, "__aexit__"):
            await session.__aexit__(*sys.exc_info())
        else:
            session.__exit__(*sys.exc_info())

    return text


async def fetch_multiple(method, loop, url):
    coros = [fetch(method, url, loop) for _ in range(2)]
    try:
        responses = await asyncio.gather(*coros, loop=loop)
    except TypeError:
        responses = await asyncio.gather(*coros)
    return "\n".join(responses)


async def multi_fetch_handler(request):
    try:
        loop = request.loop
    except AttributeError:
        loop = request.task._loop

    responses = await fetch_multiple("get", loop, request.query["url"])
    return web.Response(text=responses, content_type="text/html")


def make_app(middlewares=None):
    app = web.Application(middlewares=middlewares)
    app.router.add_route("*", "/coro", index)
    app.router.add_route("*", "/class", HelloWorldView)
    app.router.add_route("*", "/error", error)
    app.router.add_route("*", "/known_error", KnownErrorView)
    app.router.add_route("*", "/non_500_error", non_500_error)
    app.router.add_route("*", "/raise_403", raise_403)
    app.router.add_route("*", "/raise_404", raise_404)
    app.router.add_route("*", "/hang", hang)
    app.router.add_route("*", "/background", background)
    app.router.add_route("*", "/ws", websocket_handler)
    app.router.add_route("*", "/multi_fetch", multi_fetch_handler)

    for route in app.router.routes():
        handler = route.handler

        # https://github.com/aio-libs/aiohttp-cors/blob/73510a3a9212afd1d3740fd8b2feba63a1dcfe13/aiohttp_cors/urldispatcher_router_adapter.py#L95
        # Check that this statement passes for all routes
        # This statement was causing a crash since issubclass would error on
        # function wrappers
        isinstance(handler, type) and issubclass(handler, web.View)

    return app


if __name__ == "__main__":
    web.run_app(make_app(), host="127.0.0.1")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import Normalizer
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.packages import six


@pytest.mark.parametrize(
    "compose_model_name",
    [
        "ColumnTransformer",
        "TransformedTargetRegressor",
    ],
)
def test_model_methods_wrapped_in_function_trace(compose_model_name, run_compose_model):
    expected_scoped_metrics = {
        "ColumnTransformer": [
            ("Function/MLModel/Sklearn/Named/ColumnTransformer.fit", 1),
            ("Function/MLModel/Sklearn/Named/ColumnTransformer.transform", 1),
        ],
        "TransformedTargetRegressor": [
            ("Function/MLModel/Sklearn/Named/TransformedTargetRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/TransformedTargetRegressor.predict", 1),
        ],
    }

    expected_transaction_name = (
        "test_compose_models:test_model_methods_wrapped_in_function_trace.<locals>._test"
        if six.PY3
        else "test_compose_models:_test"
    )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[compose_model_name],
        rollup_metrics=expected_scoped_metrics[compose_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_compose_model(compose_model_name)

    _test()


@pytest.fixture
def run_compose_model():
    def _run(compose_model_name):
        import numpy as np
        import sklearn.compose

        if compose_model_name == "TransformedTargetRegressor":
            kwargs = {"regressor": LinearRegression()}
            X = np.arange(4).reshape(-1, 1)
            y = np.exp(2 * X).ravel()
        else:
            X = [[0.0, 1.0, 2.0, 2.0], [1.0, 1.0, 0.0, 1.0]]
            y = None
            kwargs = {
                "transformers": [
                    ("norm1", Normalizer(norm="l1"), [0, 1]),
                    ("norm2", Normalizer(norm="l1"), slice(2, 4)),
                ]
            }

        clf = getattr(sklearn.compose, compose_model_name)(**kwargs)

        model = clf.fit(X, y)
        if hasattr(model, "predict"):
            model.predict(X)
        if hasattr(model, "transform"):
            model.transform(X)

        return model

    return _run



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
import sys

import six
import webtest
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_custom_parameters import (
    validate_custom_parameters,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)

from newrelic.api.application import application_settings
from newrelic.api.transaction import (
    add_custom_attribute,
    disable_browser_autorum,
    get_browser_timing_header,
)
from newrelic.api.web_transaction import web_transaction
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.encoding_utils import deobfuscate

_runtime_error_name = RuntimeError.__module__ + ":" + RuntimeError.__name__


@wsgi_application()
def target_wsgi_application_manual_rum(environ, start_response):
    status = "200 OK"

    text = "<html><head>%s</head><body><p>RESPONSE</p></body></html>"

    output = (text % get_browser_timing_header()).encode("UTF-8")

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    return [output]


target_application_manual_rum = webtest.TestApp(target_wsgi_application_manual_rum)

_test_header_attributes = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": False,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_header_attributes)
def test_header_attributes():
    settings = application_settings()

    assert settings.browser_monitoring.enabled

    assert settings.browser_key
    assert settings.browser_monitoring.loader_version
    assert settings.js_agent_loader
    assert isinstance(settings.js_agent_file, six.string_types)
    assert settings.beacon
    assert settings.error_beacon

    token = "0123456789ABCDEF"  # nosec
    headers = {"Cookie": "NRAGENT=tk=%s" % token}

    response = target_application_manual_rum.get("/", headers=headers)

    header = response.html.html.head.script.string
    content = response.html.html.body.p.string

    # Validate actual body content.

    assert content == "RESPONSE"

    # Validate the insertion of RUM header.

    assert header.find("NREUM HEADER") != -1

    # Now validate the various fields of the header. The fields are
    # held by a JSON dictionary.

    data = json.loads(header.split("NREUM.info=")[1].split(";\n")[0])

    assert data["licenseKey"] == settings.browser_key
    assert data["applicationID"] == settings.application_id

    assert data["agent"] == settings.js_agent_file
    assert data["beacon"] == settings.beacon
    assert data["errorBeacon"] == settings.error_beacon

    assert data["applicationTime"] >= 0
    assert data["queueTime"] >= 0

    obfuscation_key = settings.license_key[:13]

    type_transaction_data = unicode if six.PY2 else str  # noqa: F821
    assert isinstance(data["transactionName"], type_transaction_data)

    txn_name = deobfuscate(data["transactionName"], obfuscation_key)

    assert txn_name == "WebTransaction/Uri/"

    assert "atts" not in data


_test_rum_ssl_for_http_is_none = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": False,
    "browser_monitoring.ssl_for_http": None,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_rum_ssl_for_http_is_none)
def test_ssl_for_http_is_none():
    settings = application_settings()

    assert settings.browser_monitoring.ssl_for_http is None

    response = target_application_manual_rum.get("/")
    header = response.html.html.head.script.string
    data = json.loads(header.split("NREUM.info=")[1].split(";\n")[0])

    assert "sslForHttp" not in data


_test_rum_ssl_for_http_is_true = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": False,
    "browser_monitoring.ssl_for_http": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_rum_ssl_for_http_is_true)
def test_ssl_for_http_is_true():
    settings = application_settings()

    assert settings.browser_monitoring.ssl_for_http is True

    response = target_application_manual_rum.get("/")
    header = response.html.html.head.script.string
    data = json.loads(header.split("NREUM.info=")[1].split(";\n")[0])

    assert data["sslForHttp"] is True


_test_rum_ssl_for_http_is_false = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": False,
    "browser_monitoring.ssl_for_http": False,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_rum_ssl_for_http_is_false)
def test_ssl_for_http_is_false():
    settings = application_settings()

    assert settings.browser_monitoring.ssl_for_http is False

    response = target_application_manual_rum.get("/")
    header = response.html.html.head.script.string
    data = json.loads(header.split("NREUM.info=")[1].split(";\n")[0])

    assert data["sslForHttp"] is False


@wsgi_application()
def target_wsgi_application_yield_single_no_head(environ, start_response):
    status = "200 OK"

    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    yield output


target_application_yield_single_no_head = webtest.TestApp(target_wsgi_application_yield_single_no_head)

_test_html_insertion_yield_single_no_head_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_yield_single_no_head_settings)
def test_html_insertion_yield_single_no_head():
    response = target_application_yield_single_no_head.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")


@wsgi_application()
def target_wsgi_application_yield_multi_no_head(environ, start_response):
    status = "200 OK"

    output = [b"<html>", b"<body><p>RESPONSE</p></body></html>"]

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(b"".join(output))))]
    start_response(status, response_headers)

    for data in output:
        yield data


target_application_yield_multi_no_head = webtest.TestApp(target_wsgi_application_yield_multi_no_head)

_test_html_insertion_yield_multi_no_head_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_yield_multi_no_head_settings)
def test_html_insertion_yield_multi_no_head():
    response = target_application_yield_multi_no_head.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")


@wsgi_application()
def target_wsgi_application_unnamed_attachment_header(environ, start_response):
    status = "200 OK"

    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [
        ("Content-Type", "text/html; charset=utf-8"),
        ("Content-Length", str(len(output))),
        ("Content-Disposition", "attachment"),
    ]
    start_response(status, response_headers)

    yield output


target_application_unnamed_attachment_header = webtest.TestApp(target_wsgi_application_unnamed_attachment_header)

_test_html_insertion_unnamed_attachment_header_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_unnamed_attachment_header_settings)
def test_html_insertion_unnamed_attachment_header():
    response = target_application_unnamed_attachment_header.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers
    assert "Content-Disposition" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@wsgi_application()
def target_wsgi_application_named_attachment_header(environ, start_response):
    status = "200 OK"

    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [
        ("Content-Type", "text/html; charset=utf-8"),
        ("Content-Length", str(len(output))),
        ("Content-Disposition", 'Attachment; filename="X"'),
    ]
    start_response(status, response_headers)

    yield output


target_application_named_attachment_header = webtest.TestApp(target_wsgi_application_named_attachment_header)

_test_html_insertion_named_attachment_header_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_named_attachment_header_settings)
def test_html_insertion_named_attachment_header():
    response = target_application_named_attachment_header.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers
    assert "Content-Disposition" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@wsgi_application()
def target_wsgi_application_inline_attachment_header(environ, start_response):
    status = "200 OK"

    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [
        ("Content-Type", "text/html; charset=utf-8"),
        ("Content-Length", str(len(output))),
        ("Content-Disposition", 'inline; filename="attachment"'),
    ]
    start_response(status, response_headers)

    yield output


target_application_inline_attachment_header = webtest.TestApp(target_wsgi_application_inline_attachment_header)

_test_html_insertion_inline_attachment_header_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_inline_attachment_header_settings)
def test_html_insertion_inline_attachment_header():
    response = target_application_inline_attachment_header.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers
    assert "Content-Disposition" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")


@wsgi_application()
def target_wsgi_application_empty_list(environ, start_response):
    status = "200 OK"

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", "0")]
    start_response(status, response_headers)

    return []


target_application_empty_list = webtest.TestApp(target_wsgi_application_empty_list)

_test_html_insertion_empty_list_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_empty_list_settings)
def test_html_insertion_empty_list():
    response = target_application_empty_list.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])

    assert len(response.body) == 0


@wsgi_application()
def target_wsgi_application_single_empty_string(environ, start_response):
    status = "200 OK"

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", "0")]
    start_response(status, response_headers)

    return [""]


target_application_single_empty_string = webtest.TestApp(target_wsgi_application_single_empty_string)

_test_html_insertion_single_empty_string_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_single_empty_string_settings)
def test_html_insertion_single_empty_string():
    response = target_application_single_empty_string.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])

    assert len(response.body) == 0


@wsgi_application()
def target_wsgi_application_multiple_empty_string(environ, start_response):
    status = "200 OK"

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", "0")]
    start_response(status, response_headers)

    return ["", ""]


target_application_multiple_empty_string = webtest.TestApp(target_wsgi_application_multiple_empty_string)

_test_html_insertion_multiple_empty_string_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_multiple_empty_string_settings)
def test_html_insertion_multiple_empty_string():
    response = target_application_multiple_empty_string.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])

    assert len(response.body) == 0


@wsgi_application()
def target_wsgi_application_single_large_prelude(environ, start_response):
    status = "200 OK"

    output = [64 * 1024 * b" " + b"<body></body>"]

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(b"".join(output))))]
    start_response(status, response_headers)

    return output


target_application_single_large_prelude = webtest.TestApp(target_wsgi_application_single_large_prelude)

_test_html_insertion_single_large_prelude_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_single_large_prelude_settings)
def test_html_insertion_single_large_prelude():
    response = target_application_single_large_prelude.get("/", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])

    output = [32 * 1024 * b" ", 32 * 1024 * b" ", b"<body></body>"]

    assert len(response.body) == len(b"".join(output))


@wsgi_application()
def target_wsgi_application_multi_large_prelude(environ, start_response):
    status = "200 OK"

    output = [32 * 1024 * b" ", 32 * 1024 * b" ", b"<body></body>"]

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(b"".join(output))))]
    start_response(status, response_headers)

    return output


target_application_multi_large_prelude = webtest.TestApp(target_wsgi_application_multi_large_prelude)

_test_html_insertion_multi_large_prelude_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_multi_large_prelude_settings)
def test_html_insertion_multi_large_prelude():
    response = target_application_multi_large_prelude.get("/", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])

    output = [32 * 1024 * b" ", 32 * 1024 * b" ", b"<body></body>"]

    assert len(response.body) == len(b"".join(output))


@wsgi_application()
def target_wsgi_application_yield_before_start(environ, start_response):
    status = "200 OK"

    # Ambiguous whether yield an empty string before calling
    # start_response() is legal. Various WSGI servers allow it
    # We have to disable WebTest lint check to get this to run.

    yield b""

    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    yield output


target_application_yield_before_start = webtest.TestApp(target_wsgi_application_yield_before_start, lint=False)

_test_html_insertion_yield_before_start_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_yield_before_start_settings)
def test_html_insertion_yield_before_start():
    response = target_application_yield_before_start.get("/", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")


@wsgi_application()
def target_wsgi_application_start_yield_start(environ, start_response):
    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]

    start_response("200 OK", response_headers)

    yield ""

    try:
        start_response(status, response_headers)  # noqa: F821
    except Exception:
        start_response("500 Error", response_headers, sys.exc_info())

    yield output


target_application_start_yield_start = webtest.TestApp(target_wsgi_application_start_yield_start)

_test_html_insertion_start_yield_start_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_start_yield_start_settings)
def test_html_insertion_start_yield_start():
    response = target_application_start_yield_start.get("/", status=500)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    response.mustcontain("NREUM HEADER", "NREUM.info")


@wsgi_application()
def target_wsgi_application_invalid_content_length(environ, start_response):
    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", "XXX")]

    start_response("200 OK", response_headers)

    yield output


target_application_invalid_content_length = webtest.TestApp(target_wsgi_application_invalid_content_length)

_test_html_insertion_invalid_content_length_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_invalid_content_length_settings)
def test_html_insertion_invalid_content_length():
    response = target_application_invalid_content_length.get("/", status=200)

    # This is relying on WebTest not validating the
    # value of the Content-Length response header
    # and just passing it through as is.

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    assert response.headers["Content-Length"] == "XXX"

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@wsgi_application()
def target_wsgi_application_content_encoding(environ, start_response):
    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [
        ("Content-Type", "text/html; charset=utf-8"),
        ("Content-Length", str(len(output))),
        ("Content-Encoding", "identity"),
    ]

    start_response("200 OK", response_headers)

    yield output


target_application_content_encoding = webtest.TestApp(target_wsgi_application_content_encoding)

_test_html_insertion_content_encoding_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_content_encoding_settings)
def test_html_insertion_content_encoding():
    response = target_application_content_encoding.get("/", status=200)

    # Technically 'identity' should not be used in Content-Encoding
    # but clients will still accept it. Use this fact to disable auto
    # RUM for this test. Other option is to compress the response
    # and use 'gzip'.

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    assert response.headers["Content-Encoding"] == "identity"

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@wsgi_application()
def target_wsgi_application_no_content_type(environ, start_response):
    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [("Content-Length", str(len(output)))]

    start_response("200 OK", response_headers)

    yield output


target_application_no_content_type = webtest.TestApp(target_wsgi_application_no_content_type, lint=False)

_test_html_insertion_no_content_type_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_no_content_type_settings)
def test_html_insertion_no_content_type():
    response = target_application_no_content_type.get("/", status=200)

    assert "Content-Type" not in response.headers
    assert "Content-Length" in response.headers

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@wsgi_application()
def target_wsgi_application_plain_text(environ, start_response):
    output = b"RESPONSE"

    response_headers = [("Content-Type", "text/plain"), ("Content-Length", str(len(output)))]

    start_response("200 OK", response_headers)

    yield output


target_application_plain_text = webtest.TestApp(target_wsgi_application_plain_text)

_test_html_insertion_plain_text_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_plain_text_settings)
def test_html_insertion_plain_text():
    response = target_application_plain_text.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@wsgi_application()
def target_wsgi_application_write_callback(environ, start_response):
    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [("Content-Type", "text/html"), ("Content-Length", str(len(output)))]

    write = start_response("200 OK", response_headers)

    write(output)

    return []


target_application_write_callback = webtest.TestApp(target_wsgi_application_write_callback)

_test_html_insertion_write_callback_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_write_callback_settings)
def test_html_insertion_write_callback():
    response = target_application_write_callback.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@wsgi_application()
def target_wsgi_application_yield_before_write(environ, start_response):
    output = [b"<html>", b"<body><p>RESPONSE</p></body></html>"]

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(b"".join(output))))]

    write = start_response("200 OK", response_headers)

    # Technically this is in violation of the WSGI specification
    # if that write() should always be before yields.

    yield output.pop(0)

    write(output.pop(0))


target_application_yield_before_write = webtest.TestApp(target_wsgi_application_yield_before_write)

_test_html_insertion_yield_before_write_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_yield_before_write_settings)
def test_html_insertion_yield_before_write():
    response = target_application_yield_before_write.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])

    expected = b"<html><body><p>RESPONSE</p></body></html>"

    assert response.body == expected


@wsgi_application()
def target_wsgi_application_write_before_yield(environ, start_response):
    output = [b"<html>", b"<body><p>RESPONSE</p></body></html>"]

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(b"".join(output))))]

    write = start_response("200 OK", response_headers)

    write(output.pop(0))

    yield output.pop(0)


target_application_write_before_yield = webtest.TestApp(target_wsgi_application_write_before_yield)

_test_html_insertion_write_before_yield_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_write_before_yield_settings)
def test_html_insertion_write_before_yield():
    response = target_application_write_before_yield.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])

    expected = b"<html><body><p>RESPONSE</p></body></html>"

    assert response.body == expected


@wsgi_application()
def target_wsgi_application_param_on_close(environ, start_response):
    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]

    start_response("200 OK", response_headers)

    try:
        yield output

    finally:
        add_custom_attribute("key", "value")


target_application_param_on_close = webtest.TestApp(target_wsgi_application_param_on_close)

_test_html_insertion_param_on_close_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_param_on_close_settings)
@validate_custom_parameters(required_params=[("key", "value")])
def test_html_insertion_param_on_close():
    response = target_application_param_on_close.get("/", status=200)

    response.mustcontain("NREUM HEADER", "NREUM.info")


@wsgi_application()
def target_wsgi_application_param_on_error(environ, start_response):
    output = b"<html><body><p>RESPONSE</p></body></html>"

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]

    start_response("200 OK", response_headers)

    try:
        raise RuntimeError("ERROR")

        yield output

    finally:
        add_custom_attribute("key", "value")


target_application_param_on_error = webtest.TestApp(target_wsgi_application_param_on_error)

_test_html_insertion_param_on_error_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_param_on_error_settings)
@validate_transaction_errors(errors=[_runtime_error_name])
@validate_custom_parameters(required_params=[("key", "value")])
def test_html_insertion_param_on_error():
    try:
        response = target_application_param_on_error.get("/", status=500)

    except RuntimeError:
        pass


@wsgi_application()
def target_wsgi_application_disable_autorum_via_api(environ, start_response):
    status = "200 OK"

    output = b"<html><body><p>RESPONSE</p></body></html>"

    disable_browser_autorum()

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    yield output


target_application_disable_autorum_via_api = webtest.TestApp(target_wsgi_application_disable_autorum_via_api)

_test_html_insertion_disable_autorum_via_api_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_disable_autorum_via_api_settings)
def test_html_insertion_disable_autorum_via_api():
    response = target_application_disable_autorum_via_api.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@wsgi_application()
def target_wsgi_application_manual_rum_insertion(environ, start_response):
    status = "200 OK"

    output = b"<html><body><p>RESPONSE</p></body></html>"

    header = get_browser_timing_header()
    header = get_browser_timing_header()

    assert header == ""

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    yield output


target_application_manual_rum_insertion = webtest.TestApp(target_wsgi_application_manual_rum_insertion)

_test_html_insertion_manual_rum_insertion_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_manual_rum_insertion_settings)
def test_html_insertion_manual_rum_insertion():
    response = target_application_manual_rum_insertion.get("/", status=200)

    assert "Content-Type" in response.headers
    assert "Content-Length" in response.headers

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


_test_get_browser_timing_snippet_with_nonces = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": False,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}
_test_get_browser_timing_snippet_with_nonces_rum_info_re = re.compile(r"NREUM\.info={[^}]*}")


@override_application_settings(_test_get_browser_timing_snippet_with_nonces)
@web_transaction(
    scheme="http", host="127.0.0.1", port=80, request_method="GET", request_path="/", query_string=None, headers={}
)
def test_get_browser_timing_snippet_with_nonces():
    header = get_browser_timing_header("NONCE")

    header = _test_get_browser_timing_snippet_with_nonces_rum_info_re.sub("NREUM.info={}", header)
    assert (
        header
        == '<script type="text/javascript" nonce="NONCE">window.NREUM||(NREUM={});NREUM.info={};\n<!-- NREUM HEADER --></script>'
    )


@override_application_settings(_test_get_browser_timing_snippet_with_nonces)
@web_transaction(
    scheme="http", host="127.0.0.1", port=80, request_method="GET", request_path="/", query_string=None, headers={}
)
def test_get_browser_timing_snippet_without_nonces():
    header = get_browser_timing_header()

    header = _test_get_browser_timing_snippet_with_nonces_rum_info_re.sub("NREUM.info={}", header)
    assert (
        header
        == '<script type="text/javascript">window.NREUM||(NREUM={});NREUM.info={};\n<!-- NREUM HEADER --></script>'
    )



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3

import pymongo
from testing_support.db_settings import mongodb_settings
from testing_support.validators.validate_database_duration import (
    validate_database_duration,
)
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from newrelic.api.background_task import background_task
from newrelic.packages import six

DB_SETTINGS = mongodb_settings()[0]
MONGODB_HOST = DB_SETTINGS["host"]
MONGODB_PORT = DB_SETTINGS["port"]
MONGODB_COLLECTION = DB_SETTINGS["collection"]


def _exercise_mongo_v3(db):
    db[MONGODB_COLLECTION].save({"x": 10})
    db[MONGODB_COLLECTION].save({"x": 8})
    db[MONGODB_COLLECTION].save({"x": 11})

    db[MONGODB_COLLECTION].find_one()

    for item in db[MONGODB_COLLECTION].find():
        item["x"]

    db[MONGODB_COLLECTION].create_index("x")

    for item in db[MONGODB_COLLECTION].find().sort("x", pymongo.ASCENDING):
        item["x"]

    [item["x"] for item in db[MONGODB_COLLECTION].find().limit(2).skip(1)]

    db[MONGODB_COLLECTION].initialize_unordered_bulk_op()
    db[MONGODB_COLLECTION].initialize_ordered_bulk_op()
    db[MONGODB_COLLECTION].parallel_scan(1)

    db[MONGODB_COLLECTION].bulk_write([pymongo.InsertOne({"x": 1})])
    db[MONGODB_COLLECTION].insert_one({"x": 300})
    db[MONGODB_COLLECTION].insert_many([{"x": 1} for i in range(20, 25)])
    db[MONGODB_COLLECTION].replace_one({"x": 1}, {"x": 2})
    db[MONGODB_COLLECTION].update_one({"x": 1}, {"$inc": {"x": 3}})
    db[MONGODB_COLLECTION].update_many({"x": 1}, {"$inc": {"x": 3}})
    db[MONGODB_COLLECTION].delete_one({"x": 4})
    db[MONGODB_COLLECTION].delete_many({"x": 4})
    db[MONGODB_COLLECTION].find_raw_batches()
    db[MONGODB_COLLECTION].create_indexes([pymongo.IndexModel([("x", pymongo.DESCENDING)])])
    db[MONGODB_COLLECTION].list_indexes()
    db[MONGODB_COLLECTION].aggregate([])
    db[MONGODB_COLLECTION].aggregate_raw_batches([])
    db[MONGODB_COLLECTION].find_one_and_delete({"x": 10})
    db[MONGODB_COLLECTION].find_one_and_replace({"x": 300}, {"x": 301})
    db[MONGODB_COLLECTION].find_one_and_update({"x": 301}, {"$inc": {"x": 300}})


def _exercise_mongo_v4(db):
    db[MONGODB_COLLECTION].insert_one({"x": 10})
    db[MONGODB_COLLECTION].insert_one({"x": 8})
    db[MONGODB_COLLECTION].insert_one({"x": 11})

    db[MONGODB_COLLECTION].find_one()

    for item in db[MONGODB_COLLECTION].find():
        item["x"]

    db[MONGODB_COLLECTION].create_index("x")

    for item in db[MONGODB_COLLECTION].find().sort("x", pymongo.ASCENDING):
        item["x"]

    [item["x"] for item in db[MONGODB_COLLECTION].find().limit(2).skip(1)]

    db[MONGODB_COLLECTION].bulk_write([pymongo.InsertOne({"x": 1})])
    db[MONGODB_COLLECTION].insert_one({"x": 300})
    db[MONGODB_COLLECTION].insert_many([{"x": 1} for i in range(20, 25)])
    db[MONGODB_COLLECTION].replace_one({"x": 1}, {"x": 2})
    db[MONGODB_COLLECTION].update_one({"x": 1}, {"$inc": {"x": 3}})
    db[MONGODB_COLLECTION].update_many({"x": 1}, {"$inc": {"x": 3}})
    db[MONGODB_COLLECTION].delete_one({"x": 4})
    db[MONGODB_COLLECTION].delete_many({"x": 4})
    db[MONGODB_COLLECTION].find_raw_batches()
    db[MONGODB_COLLECTION].create_indexes([pymongo.IndexModel([("x", pymongo.DESCENDING)])])
    db[MONGODB_COLLECTION].list_indexes()
    db[MONGODB_COLLECTION].aggregate([])
    db[MONGODB_COLLECTION].aggregate_raw_batches([])
    db[MONGODB_COLLECTION].find_one_and_delete({"x": 10})
    db[MONGODB_COLLECTION].find_one_and_replace({"x": 300}, {"x": 301})
    db[MONGODB_COLLECTION].find_one_and_update({"x": 301}, {"$inc": {"x": 300}})


def _exercise_mongo(db):
    if pymongo.version_tuple < (4, 0):
        _exercise_mongo_v3(db)
    else:
        _exercise_mongo_v4(db)


# Common Metrics for tests that use _exercise_mongo().


_test_pymongo_scoped_metrics_v3 = [
    ("Function/pymongo.mongo_client:MongoClient.__init__", 1),
    ("Datastore/statement/MongoDB/%s/create_index" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find" % MONGODB_COLLECTION, 3),
    ("Datastore/statement/MongoDB/%s/find_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/save" % MONGODB_COLLECTION, 3),
    ("Datastore/statement/MongoDB/%s" % MONGODB_COLLECTION + "/initialize_unordered_bulk_op", 1),
    ("Datastore/statement/MongoDB/%s" % MONGODB_COLLECTION + "/initialize_ordered_bulk_op", 1),
    ("Datastore/statement/MongoDB/%s/parallel_scan" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/insert_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/bulk_write" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/insert_many" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/replace_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/update_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/delete_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/delete_many" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_raw_batches" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/create_indexes" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/list_indexes" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/aggregate" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/aggregate_raw_batches" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_delete" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_replace" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_update" % MONGODB_COLLECTION, 1),
]


_test_pymongo_scoped_metrics_v4 = [
    ("Function/pymongo.mongo_client:MongoClient.__init__", 1),
    ("Datastore/statement/MongoDB/%s/create_index" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find" % MONGODB_COLLECTION, 3),
    ("Datastore/statement/MongoDB/%s/find_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/insert_one" % MONGODB_COLLECTION, 4),
    ("Datastore/statement/MongoDB/%s/bulk_write" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/insert_many" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/replace_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/update_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/delete_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/delete_many" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_raw_batches" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/create_indexes" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/list_indexes" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/aggregate" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/aggregate_raw_batches" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_delete" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_replace" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_update" % MONGODB_COLLECTION, 1),
]

_test_pymongo_rollup_metrics_v3 = [
    ("Function/pymongo.mongo_client:MongoClient.__init__", 1),
    ("Datastore/all", 28),
    ("Datastore/allOther", 28),
    ("Datastore/MongoDB/all", 28),
    ("Datastore/MongoDB/allOther", 28),
    ("Datastore/operation/MongoDB/create_index", 1),
    ("Datastore/operation/MongoDB/find", 3),
    ("Datastore/operation/MongoDB/find_one", 1),
    ("Datastore/statement/MongoDB/%s/create_index" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find" % MONGODB_COLLECTION, 3),
    ("Datastore/statement/MongoDB/%s/find_one" % MONGODB_COLLECTION, 1),
    ("Datastore/operation/MongoDB/save", 3),
    ("Datastore/operation/MongoDB/initialize_unordered_bulk_op", 1),
    ("Datastore/operation/MongoDB/initialize_ordered_bulk_op", 1),
    ("Datastore/operation/MongoDB/parallel_scan", 1),
    ("Datastore/statement/MongoDB/%s/save" % MONGODB_COLLECTION, 3),
    (("Datastore/statement/MongoDB/%s" % MONGODB_COLLECTION + "/initialize_unordered_bulk_op"), 1),
    (("Datastore/statement/MongoDB/%s" % MONGODB_COLLECTION + "/initialize_ordered_bulk_op"), 1),
    ("Datastore/statement/MongoDB/%s/parallel_scan" % MONGODB_COLLECTION, 1),
    ("Datastore/operation/MongoDB/bulk_write", 1),
    ("Datastore/operation/MongoDB/insert_one", 1),
    ("Datastore/operation/MongoDB/insert_many", 1),
    ("Datastore/operation/MongoDB/replace_one", 1),
    ("Datastore/operation/MongoDB/update_one", 1),
    ("Datastore/operation/MongoDB/delete_one", 1),
    ("Datastore/operation/MongoDB/delete_many", 1),
    ("Datastore/operation/MongoDB/find_raw_batches", 1),
    ("Datastore/operation/MongoDB/create_indexes", 1),
    ("Datastore/operation/MongoDB/list_indexes", 1),
    ("Datastore/operation/MongoDB/aggregate", 1),
    ("Datastore/operation/MongoDB/aggregate_raw_batches", 1),
    ("Datastore/operation/MongoDB/find_one_and_delete", 1),
    ("Datastore/operation/MongoDB/find_one_and_replace", 1),
    ("Datastore/operation/MongoDB/find_one_and_update", 1),
    ("Datastore/statement/MongoDB/%s/bulk_write" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/insert_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/insert_many" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/replace_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/update_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/delete_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/delete_many" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_raw_batches" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/create_indexes" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/list_indexes" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/aggregate" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/aggregate_raw_batches" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_delete" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_replace" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_update" % MONGODB_COLLECTION, 1),
]

_test_pymongo_rollup_metrics_v4 = [
    ("Function/pymongo.mongo_client:MongoClient.__init__", 1),
    ("Datastore/all", 25),
    ("Datastore/allOther", 25),
    ("Datastore/MongoDB/all", 25),
    ("Datastore/MongoDB/allOther", 25),
    ("Datastore/operation/MongoDB/create_index", 1),
    ("Datastore/operation/MongoDB/find", 3),
    ("Datastore/operation/MongoDB/find_one", 1),
    ("Datastore/statement/MongoDB/%s/create_index" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find" % MONGODB_COLLECTION, 3),
    ("Datastore/statement/MongoDB/%s/find_one" % MONGODB_COLLECTION, 1),
    ("Datastore/operation/MongoDB/bulk_write", 1),
    ("Datastore/operation/MongoDB/insert_one", 4),
    ("Datastore/operation/MongoDB/insert_many", 1),
    ("Datastore/operation/MongoDB/replace_one", 1),
    ("Datastore/operation/MongoDB/update_one", 1),
    ("Datastore/operation/MongoDB/delete_one", 1),
    ("Datastore/operation/MongoDB/delete_many", 1),
    ("Datastore/operation/MongoDB/find_raw_batches", 1),
    ("Datastore/operation/MongoDB/create_indexes", 1),
    ("Datastore/operation/MongoDB/list_indexes", 1),
    ("Datastore/operation/MongoDB/aggregate", 1),
    ("Datastore/operation/MongoDB/aggregate_raw_batches", 1),
    ("Datastore/operation/MongoDB/find_one_and_delete", 1),
    ("Datastore/operation/MongoDB/find_one_and_replace", 1),
    ("Datastore/operation/MongoDB/find_one_and_update", 1),
    ("Datastore/statement/MongoDB/%s/bulk_write" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/insert_one" % MONGODB_COLLECTION, 4),
    ("Datastore/statement/MongoDB/%s/insert_many" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/replace_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/update_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/delete_one" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/delete_many" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_raw_batches" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/create_indexes" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/list_indexes" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/aggregate" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/aggregate_raw_batches" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_delete" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_replace" % MONGODB_COLLECTION, 1),
    ("Datastore/statement/MongoDB/%s/find_one_and_update" % MONGODB_COLLECTION, 1),
]


def test_mongodb_client_operation():
    if pymongo.version_tuple < (4, 0):
        _test_pymongo_client_scoped_metrics = _test_pymongo_scoped_metrics_v3
        _test_pymongo_client_rollup_metrics = _test_pymongo_rollup_metrics_v3
    else:
        _test_pymongo_client_scoped_metrics = _test_pymongo_scoped_metrics_v4
        _test_pymongo_client_rollup_metrics = _test_pymongo_rollup_metrics_v4

    txn_name = "test_pymongo:test_mongodb_client_operation.<locals>._test" if six.PY3 else "test_pymongo:_test"

    @validate_transaction_errors(errors=[])
    @validate_transaction_metrics(
        txn_name,
        scoped_metrics=_test_pymongo_client_scoped_metrics,
        rollup_metrics=_test_pymongo_client_rollup_metrics,
        background_task=True,
    )
    @background_task()
    def _test():
        client = pymongo.MongoClient(MONGODB_HOST, MONGODB_PORT)
        db = client.test
        _exercise_mongo(db)

    _test()


@validate_database_duration()
@background_task()
def test_mongodb_database_duration():
    client = pymongo.MongoClient(MONGODB_HOST, MONGODB_PORT)
    db = client.test
    _exercise_mongo(db)


@validate_database_duration()
@background_task()
def test_mongodb_and_sqlite_database_duration():

    # Make mongodb queries

    client = pymongo.MongoClient(MONGODB_HOST, MONGODB_PORT)
    db = client.test
    _exercise_mongo(db)

    # Make sqlite queries

    conn = sqlite3.connect(":memory:")
    cur = conn.cursor()

    cur.execute("CREATE TABLE contacts (name text, age int)")
    cur.execute("INSERT INTO contacts VALUES ('Bob', 22)")

    conn.commit()
    conn.close()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import boto3
import moto
import pytest
from testing_support.fixtures import dt_enabled
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_segment_params import (
    validate_tt_segment_params,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

MOTO_VERSION = get_package_version_tuple("moto")

# patch earlier versions of moto to support py37
if sys.version_info >= (3, 7) and MOTO_VERSION <= (1, 3, 1):
    import re

    moto.packages.responses.responses.re._pattern_type = re.Pattern

AWS_ACCESS_KEY_ID = "AAAAAAAAAAAACCESSKEY"
AWS_SECRET_ACCESS_KEY = "AAAAAASECRETKEY"  # nosec (This is fine for testing purposes)
AWS_REGION_NAME = "us-east-1"
SNS_URL = "sns-us-east-1.amazonaws.com"
TOPIC = "arn:aws:sns:us-east-1:123456789012:some-topic"
sns_metrics = [("MessageBroker/SNS/Topic" "/Produce/Named/%s" % TOPIC, 1)]
sns_metrics_phone = [("MessageBroker/SNS/Topic" "/Produce/Named/PhoneNumber", 1)]


@dt_enabled
@validate_span_events(expected_agents=("aws.requestId",), count=2)
@validate_span_events(exact_agents={"aws.operation": "CreateTopic"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "Publish"}, count=1)
@validate_tt_segment_params(present_params=("aws.requestId",))
@pytest.mark.parametrize("topic_argument", ("TopicArn", "TargetArn"))
@validate_transaction_metrics(
    "test_boto3_sns:test_publish_to_sns_topic",
    scoped_metrics=sns_metrics,
    rollup_metrics=sns_metrics,
    background_task=True,
)
@background_task()
@moto.mock_sns
def test_publish_to_sns_topic(topic_argument):
    conn = boto3.client(
        "sns",
        aws_access_key_id=AWS_ACCESS_KEY_ID,
        aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
        region_name=AWS_REGION_NAME,
    )

    topic_arn = conn.create_topic(Name="some-topic")["TopicArn"]

    kwargs = {topic_argument: topic_arn}
    published_message = conn.publish(Message="my msg", **kwargs)
    assert "MessageId" in published_message


@dt_enabled
@validate_span_events(expected_agents=("aws.requestId",), count=3)
@validate_span_events(exact_agents={"aws.operation": "CreateTopic"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "Subscribe"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "Publish"}, count=1)
@validate_tt_segment_params(present_params=("aws.requestId",))
@validate_transaction_metrics(
    "test_boto3_sns:test_publish_to_sns_phone",
    scoped_metrics=sns_metrics_phone,
    rollup_metrics=sns_metrics_phone,
    background_task=True,
)
@background_task()
@moto.mock_sns
def test_publish_to_sns_phone():
    conn = boto3.client(
        "sns",
        aws_access_key_id=AWS_ACCESS_KEY_ID,
        aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
        region_name=AWS_REGION_NAME,
    )

    topic_arn = conn.create_topic(Name="some-topic")["TopicArn"]
    conn.subscribe(TopicArn=topic_arn, Protocol="sms", Endpoint="5555555555")

    published_message = conn.publish(PhoneNumber="5555555555", Message="my msg")
    assert "MessageId" in published_message



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import json
import os

from contextlib import contextmanager

from newrelic.api.application import (application_instance as
        current_application)
from newrelic.api.background_task import BackgroundTask
from newrelic.core.rules_engine import SegmentCollapseEngine
from newrelic.core.agent import agent_instance

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
JSON_DIR = os.path.normpath(os.path.join(CURRENT_DIR, 'fixtures'))
OUTBOUD_REQUESTS = {}

_parameters_list = ['testname', 'transaction_segment_terms', 'tests']


def load_tests():
    result = []
    path = os.path.join(JSON_DIR, 'transaction_segment_terms.json')
    with open(path, 'r') as fh:
        tests = json.load(fh)

    for test in tests:
        values = tuple([test.get(param, None) for param in _parameters_list])
        result.append(values)

    return result

_parameters = ",".join(_parameters_list)

@pytest.mark.parametrize(_parameters, load_tests())
def test_transaction_segments(testname, transaction_segment_terms, tests):
    engine = SegmentCollapseEngine(transaction_segment_terms)
    for test in tests:
        assert engine.normalize(test['input'])[0] == test['expected']

@contextmanager
def segment_rules(name, rules):
    application = agent_instance().application(name)
    old_rules = application._rules_engine['segment']
    new_rules = SegmentCollapseEngine(rules) 
    application._rules_engine['segment'] = new_rules
    yield
    application._rules_engine['segment'] = old_rules

@pytest.mark.parametrize(_parameters, load_tests())
def test_transaction_freeze_path_segments(testname, transaction_segment_terms,
        tests):

    application = current_application()

    # We can't check all possibilites by doing things via the transaction
    # as it not possible to set up a metric path of only one segment.

    with segment_rules(application.name, transaction_segment_terms):
        for test in tests:
            segments = test['input'].split()
            if len(segments) < 2:
                continue

            ttype = segments[0]
            group = '/'.join(segments[1:2])
            name = '/'.join(segments[2:])

            with BackgroundTask(application, name, group) as transaction:
                transaction.background_task = (ttype == 'OtherTransaction')

            assert transaction.path == test['expected']



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform

from newrelic.api.application import application_settings
from newrelic.api.background_task import background_task
from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_log_event_count import validate_log_event_count
from testing_support.validators.validate_log_event_count_outside_transaction import validate_log_event_count_outside_transaction


def get_metadata_string(log_message, is_txn):
    host = platform.uname()[1]
    assert host
    entity_guid = application_settings().entity_guid
    if is_txn:
        metadata_string = "".join(('NR-LINKING|', entity_guid, '|', host, '|abcdefgh12345678|abcdefgh|Python%20Agent%20Test%20%28logger_structlog%29|'))
    else:
        metadata_string = "".join(('NR-LINKING|', entity_guid, '|', host, '|||Python%20Agent%20Test%20%28logger_structlog%29|'))
    formatted_string = log_message + " " + metadata_string
    return formatted_string


@reset_core_stats_engine()
def test_local_log_decoration_inside_transaction(exercise_logging_single_line, structlog_caplog):
    @validate_log_event_count(1)
    @background_task()
    def test():
        exercise_logging_single_line()
        assert get_metadata_string('A', True) in structlog_caplog.caplog[0]

    test()


@reset_core_stats_engine()
def test_local_log_decoration_outside_transaction(exercise_logging_single_line, structlog_caplog):
    @validate_log_event_count_outside_transaction(1)
    def test():
        exercise_logging_single_line()
        assert get_metadata_string('A', False) in structlog_caplog.caplog[0]

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

import pytest
import webtest
from testing_support.fixtures import override_application_settings, validate_attributes
from testing_support.validators.validate_error_event_attributes import (
    validate_error_event_attributes,
)
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.transaction import current_transaction
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.encoding_utils import DistributedTracePayload
from newrelic.common.object_wrapper import transient_function_wrapper

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
JSON_DIR = os.path.normpath(os.path.join(CURRENT_DIR, "fixtures", "distributed_tracing"))

_parameters_list = [
    "account_id",
    "comment",
    "expected_metrics",
    "force_sampled_true",
    "inbound_payloads",
    "intrinsics",
    "major_version",
    "minor_version",
    "outbound_payloads",
    "raises_exception",
    "span_events_enabled",
    "test_name",
    "transport_type",
    "trusted_account_key",
    "web_transaction",
]
_parameters = ",".join(_parameters_list)


def load_tests():
    result = []
    path = os.path.join(JSON_DIR, "distributed_tracing.json")
    with open(path, "r") as fh:
        tests = json.load(fh)

    for test in tests:
        values = (test.get(param, None) for param in _parameters_list)
        param = pytest.param(*values, id=test.get("test_name"))
        result.append(param)

    return result


def override_compute_sampled(override):
    @transient_function_wrapper("newrelic.core.adaptive_sampler", "AdaptiveSampler.compute_sampled")
    def _override_compute_sampled(wrapped, instance, args, kwargs):
        if override:
            return True
        return wrapped(*args, **kwargs)

    return _override_compute_sampled


def assert_payload(payload, payload_assertions, major_version, minor_version):
    assert payload

    # flatten payload so it matches the test:
    #   payload['d']['ac'] -> payload['d.ac']
    d = payload.pop("d")
    for key, value in d.items():
        payload["d.%s" % key] = value

    for expected in payload_assertions.get("expected", []):
        assert expected in payload

    for unexpected in payload_assertions.get("unexpected", []):
        assert unexpected not in payload

    for key, value in payload_assertions.get("exact", {}).items():
        assert key in payload
        if isinstance(value, list):
            value = tuple(value)
        assert payload[key] == value

    assert payload["v"][0] == major_version
    assert payload["v"][1] == minor_version


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status = "200 OK"
    output = b"hello world"
    response_headers = [("Content-type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]

    txn = current_transaction()
    txn.set_transaction_name(test_settings["test_name"])

    if not test_settings["web_transaction"]:
        txn.background_task = True

    if test_settings["raises_exception"]:
        try:
            1 / 0
        except ZeroDivisionError:
            txn.notice_error()

    extra_inbound_payloads = test_settings["extra_inbound_payloads"]
    for payload, expected_result in extra_inbound_payloads:
        result = txn.accept_distributed_trace_payload(payload, test_settings["transport_type"])
        assert result is expected_result

    outbound_payloads = test_settings["outbound_payloads"]
    if outbound_payloads:
        for payload_assertions in outbound_payloads:
            payload = txn._create_distributed_trace_payload()
            assert_payload(payload, payload_assertions, test_settings["major_version"], test_settings["minor_version"])

    start_response(status, response_headers)
    return [output]


test_application = webtest.TestApp(target_wsgi_application)


@pytest.mark.parametrize(_parameters, load_tests())
def test_distributed_tracing(
    account_id,
    comment,
    expected_metrics,
    force_sampled_true,
    inbound_payloads,
    intrinsics,
    major_version,
    minor_version,
    outbound_payloads,
    raises_exception,
    span_events_enabled,
    test_name,
    transport_type,
    trusted_account_key,
    web_transaction,
):

    extra_inbound_payloads = []
    if transport_type != "HTTP":
        # Since wsgi_application calls accept_distributed_trace_payload
        # automatically with transport_type='HTTP', we must defer this call
        # until we can specify the transport type.
        extra_inbound_payloads.append((inbound_payloads.pop(), True))
    elif not inbound_payloads:
        # In order to assert that accept_distributed_trace_payload returns
        # False in this instance, we defer.
        extra_inbound_payloads.append((inbound_payloads, False))
    elif len(inbound_payloads) > 1:
        for payload in inbound_payloads[1:]:
            extra_inbound_payloads.append((payload, False))

    global test_settings
    test_settings = {
        "test_name": test_name,
        "web_transaction": web_transaction,
        "raises_exception": raises_exception,
        "extra_inbound_payloads": extra_inbound_payloads,
        "outbound_payloads": outbound_payloads,
        "transport_type": transport_type,
        "major_version": major_version,
        "minor_version": minor_version,
    }

    override_settings = {
        "distributed_tracing.enabled": True,
        "span_events.enabled": span_events_enabled,
        "account_id": account_id,
        "trusted_account_key": trusted_account_key,
    }

    common_required = intrinsics["common"]["expected"]
    common_forgone = intrinsics["common"]["unexpected"]
    common_exact = intrinsics["common"].get("exact", {})

    txn_intrinsics = intrinsics.get("Transaction", {})
    txn_event_required = {"agent": [], "user": [], "intrinsic": txn_intrinsics.get("expected", [])}
    txn_event_required["intrinsic"].extend(common_required)
    txn_event_forgone = {"agent": [], "user": [], "intrinsic": txn_intrinsics.get("unexpected", [])}
    txn_event_forgone["intrinsic"].extend(common_forgone)
    txn_event_exact = {"agent": {}, "user": {}, "intrinsic": txn_intrinsics.get("exact", {})}
    txn_event_exact["intrinsic"].update(common_exact)

    headers = {}
    if inbound_payloads:
        payload = DistributedTracePayload(inbound_payloads[0])
        headers["newrelic"] = payload.http_safe()

    @validate_transaction_metrics(test_name, rollup_metrics=expected_metrics, background_task=not web_transaction)
    @validate_transaction_event_attributes(txn_event_required, txn_event_forgone, txn_event_exact)
    @validate_attributes("intrinsic", common_required, common_forgone)
    @override_compute_sampled(force_sampled_true)
    @override_application_settings(override_settings)
    def _test():
        response = test_application.get("/", headers=headers)
        assert "X-NewRelic-App-Data" not in response.headers

    if "Span" in intrinsics:
        span_intrinsics = intrinsics.get("Span")
        span_expected = span_intrinsics.get("expected", [])
        span_expected.extend(common_required)
        span_unexpected = span_intrinsics.get("unexpected", [])
        span_unexpected.extend(common_forgone)
        span_exact = span_intrinsics.get("exact", {})
        span_exact.update(common_exact)

        _test = validate_span_events(
            exact_intrinsics=span_exact, expected_intrinsics=span_expected, unexpected_intrinsics=span_unexpected
        )(_test)
    elif not span_events_enabled:
        _test = validate_span_events(count=0)(_test)

    if raises_exception:
        error_event_required = {"agent": [], "user": [], "intrinsic": common_required}
        error_event_forgone = {"agent": [], "user": [], "intrinsic": common_forgone}
        error_event_exact = {"agent": {}, "user": {}, "intrinsic": common_exact}
        _test = validate_error_event_attributes(error_event_required, error_event_forgone, error_event_exact)(_test)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from starlette.applications import Starlette
from starlette.routing import Route
from testing_support.asgi_testing import AsgiTest

from graphene import ObjectType, String, Schema
from graphql.execution.executors.asyncio import AsyncioExecutor
from starlette.graphql import GraphQLApp


class Query(ObjectType):
    hello = String()

    def resolve_hello(self, info):
        return "Hello!"


routes = [
    Route("/async", GraphQLApp(executor_class=AsyncioExecutor, schema=Schema(query=Query))),
    Route("/sync", GraphQLApp(schema=Schema(query=Query))),
]

app = Starlette(routes=routes)
target_application = AsgiTest(app)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.encoding_utils import unpack_field
from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper
from newrelic.core.database_utils import SQLConnections


def validate_tt_segment_params(forgone_params=(), present_params=(), exact_params=None):
    exact_params = exact_params or {}
    recorded_traces = []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _extract_trace(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        # Now that transaction has been recorded, generate
        # a transaction trace

        connections = SQLConnections()
        trace_data = instance.transaction_trace_data(connections)
        # Save the recorded traces
        recorded_traces.extend(trace_data)

        return result

    @function_wrapper
    def validator(wrapped, instance, args, kwargs):
        new_wrapper = _extract_trace(wrapped)
        result = new_wrapper(*args, **kwargs)

        # Verify that traces have been recorded
        assert recorded_traces

        # Extract the first transaction trace
        transaction_trace = recorded_traces[0]
        pack_data = unpack_field(transaction_trace[4])

        # Extract the root segment from the root node
        root_segment = pack_data[0][3]

        recorded_params = {}

        def _validate_segment_params(segment):
            segment_params = segment[3]

            # Translate from the string cache
            for key, value in segment_params.items():
                if hasattr(value, "startswith") and value.startswith("`"):
                    try:
                        index = int(value[1:])
                        value = pack_data[1][index]
                    except ValueError:
                        pass
                segment_params[key] = value

            recorded_params.update(segment_params)

            for child_segment in segment[4]:
                _validate_segment_params(child_segment)

        _validate_segment_params(root_segment)

        recorded_params_set = set(recorded_params.keys())

        # Verify that the params in present params have been recorded
        present_params_set = set(present_params)
        assert recorded_params_set.issuperset(present_params_set)

        # Verify that all forgone params are omitted
        recorded_forgone_params = recorded_params_set & set(forgone_params)
        assert not recorded_forgone_params

        # Verify that all exact params are correct
        for key, value in exact_params.items():
            assert recorded_params[key] == value

        return result

    return validator



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides common utilities for interacting with OTLP protocol buffers.

The serialization implemented here attempts to use protobuf as an encoding, but falls
back to JSON when encoutering exceptions unless the content type is explicitly set in debug settings.
"""

import logging

from newrelic.common.encoding_utils import json_encode
from newrelic.core.config import global_settings
from newrelic.core.stats_engine import CountStats, TimeStats

_logger = logging.getLogger(__name__)

_settings = global_settings()
otlp_content_setting = _settings.debug.otlp_content_encoding
if not otlp_content_setting or otlp_content_setting == "protobuf":
    try:
        from newrelic.packages.opentelemetry_proto.common_pb2 import AnyValue, KeyValue
        from newrelic.packages.opentelemetry_proto.logs_pb2 import (
            LogsData,
            ResourceLogs,
            ScopeLogs,
        )
        from newrelic.packages.opentelemetry_proto.metrics_pb2 import (
            AggregationTemporality,
            Metric,
            MetricsData,
            NumberDataPoint,
            ResourceMetrics,
            ScopeMetrics,
            Sum,
            Summary,
            SummaryDataPoint,
        )
        from newrelic.packages.opentelemetry_proto.resource_pb2 import Resource

        ValueAtQuantile = SummaryDataPoint.ValueAtQuantile
        AGGREGATION_TEMPORALITY_DELTA = AggregationTemporality.AGGREGATION_TEMPORALITY_DELTA
        OTLP_CONTENT_TYPE = "application/x-protobuf"

        otlp_content_setting = "protobuf"  # Explicitly set to overwrite None values
    except Exception:
        if otlp_content_setting == "protobuf":
            raise  # Reraise exception if content type explicitly set
        # Fallback to JSON
        otlp_content_setting = "json"


if otlp_content_setting == "json":
    AnyValue = dict
    KeyValue = dict
    Metric = dict
    MetricsData = dict
    NumberDataPoint = dict
    Resource = dict
    ResourceMetrics = dict
    ScopeMetrics = dict
    Sum = dict
    Summary = dict
    SummaryDataPoint = dict
    ValueAtQuantile = dict
    ResourceLogs = dict
    ScopeLogs = dict
    LogsData = dict

    AGGREGATION_TEMPORALITY_DELTA = 1
    OTLP_CONTENT_TYPE = "application/json"


def otlp_encode(payload):
    if type(payload) is dict:  # pylint: disable=C0123
        _logger.warning(
            "Using OTLP integration while protobuf is not installed. This may result in larger payload sizes and data loss."
        )
        return json_encode(payload).encode("utf-8")
    return payload.SerializeToString()


def create_key_value(key, value):
    if isinstance(value, bool):
        return KeyValue(key=key, value=AnyValue(bool_value=value))
    elif isinstance(value, int):
        return KeyValue(key=key, value=AnyValue(int_value=value))
    elif isinstance(value, float):
        return KeyValue(key=key, value=AnyValue(double_value=value))
    elif isinstance(value, str):
        return KeyValue(key=key, value=AnyValue(string_value=value))
    # Technically AnyValue accepts array, kvlist, and bytes however, since
    # those are not valid custom attribute types according to our api spec,
    # we will not bother to support them here either.
    else:
        _logger.warning("Unsupported attribute value type %s: %s." % (key, value))


def create_key_values_from_iterable(iterable):
    if not iterable:
        return None
    elif isinstance(iterable, dict):
        iterable = iterable.items()

    # The create_key_value list may return None if the value is an unsupported type
    # so filter None values out before returning.
    return list(
        filter(
            lambda i: i is not None,
            (create_key_value(key, value) for key, value in iterable),
        )
    )


def create_resource(attributes=None):
    attributes = attributes or {"instrumentation.provider": "newrelic-opentelemetry-python-ml"}
    return Resource(attributes=create_key_values_from_iterable(attributes))


def TimeStats_to_otlp_data_point(self, start_time, end_time, attributes=None):
    data = SummaryDataPoint(
        time_unix_nano=int(end_time * 1e9),  # Time of current harvest
        start_time_unix_nano=int(start_time * 1e9),  # Time of last harvest
        attributes=attributes,
        count=int(self[0]),
        sum=float(self[1]),
        quantile_values=[
            ValueAtQuantile(quantile=0.0, value=float(self[3])),  # Min Value
            ValueAtQuantile(quantile=1.0, value=float(self[4])),  # Max Value
        ],
    )
    return data


def CountStats_to_otlp_data_point(self, start_time, end_time, attributes=None):
    data = NumberDataPoint(
        time_unix_nano=int(end_time * 1e9),  # Time of current harvest
        start_time_unix_nano=int(start_time * 1e9),  # Time of last harvest
        attributes=attributes,
        as_int=int(self[0]),
    )
    return data


def stats_to_otlp_metrics(metric_data, start_time, end_time):
    """
    Generator producing protos for Summary and Sum metrics, for CountStats and TimeStats respectively.

    Individual Metric protos must be entirely one type of metric data point. For mixed metric types we have to
    separate the types and report multiple metrics, one for each type.
    """
    for name, metric_container in metric_data:
        # Types are checked here using type() instead of isinstance, as CountStats is a subclass of TimeStats.
        # Imporperly checking with isinstance will lead to count metrics being encoded and reported twice.
        if any(type(metric) is CountStats for metric in metric_container.values()):  # pylint: disable=C0123
            # Metric contains Sum metric data points.
            yield Metric(
                name=name,
                sum=Sum(
                    aggregation_temporality=AGGREGATION_TEMPORALITY_DELTA,
                    is_monotonic=True,
                    data_points=[
                        CountStats_to_otlp_data_point(
                            value,
                            start_time=start_time,
                            end_time=end_time,
                            attributes=create_key_values_from_iterable(tags),
                        )
                        for tags, value in metric_container.items()
                        if type(value) is CountStats  # pylint: disable=C0123
                    ],
                ),
            )
        if any(type(metric) is TimeStats for metric in metric_container.values()):  # pylint: disable=C0123
            # Metric contains Summary metric data points.
            yield Metric(
                name=name,
                summary=Summary(
                    data_points=[
                        TimeStats_to_otlp_data_point(
                            value,
                            start_time=start_time,
                            end_time=end_time,
                            attributes=create_key_values_from_iterable(tags),
                        )
                        for tags, value in metric_container.items()
                        if type(value) is TimeStats  # pylint: disable=C0123
                    ]
                ),
            )


def encode_metric_data(metric_data, start_time, end_time, resource=None, scope=None):
    resource = resource or create_resource()
    return MetricsData(
        resource_metrics=[
            ResourceMetrics(
                resource=resource,
                scope_metrics=[
                    ScopeMetrics(
                        scope=scope,
                        metrics=list(stats_to_otlp_metrics(metric_data, start_time, end_time)),
                    )
                ],
            )
        ]
    )


def encode_ml_event_data(custom_event_data, agent_run_id):
    resource = create_resource()
    ml_events = []
    for event in custom_event_data:
        event_info, event_attrs = event
        event_attrs.update(
            {
                "real_agent_id": agent_run_id,
                "event.domain": "newrelic.ml_events",
                "event.name": event_info["type"],
            }
        )
        ml_attrs = create_key_values_from_iterable(event_attrs)
        unix_nano_timestamp = event_info["timestamp"] * 1e6
        ml_events.append(
            {
                "time_unix_nano": int(unix_nano_timestamp),
                "attributes": ml_attrs,
            }
        )

    return LogsData(resource_logs=[ResourceLogs(resource=resource, scope_logs=[ScopeLogs(log_records=ml_events)])])



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import psycopg2
import psycopg2.extensions
import psycopg2.extras
try:
    from psycopg2 import sql
except ImportError:
    sql = None

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_database_trace_inputs import validate_database_trace_inputs
from testing_support.util import instance_hostname
from utils import DB_SETTINGS

from newrelic.api.background_task import background_task


# Settings
_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
}


# Metrics
_base_scoped_metrics = (
        ('Function/psycopg2:connect', 1),
        ('Datastore/statement/Postgres/%s/select' % DB_SETTINGS['table_name'], 1),
        ('Datastore/statement/Postgres/%s/insert' % DB_SETTINGS['table_name'], 1),
        ('Datastore/statement/Postgres/%s/update' % DB_SETTINGS['table_name'], 1),
        ('Datastore/statement/Postgres/%s/delete' % DB_SETTINGS['table_name'], 1),
        ('Datastore/statement/Postgres/now/call', 1),
        ('Datastore/statement/Postgres/pg_sleep/call', 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1),
        ('Datastore/operation/Postgres/commit', 2),
        ('Datastore/operation/Postgres/rollback', 1),
)

_base_rollup_metrics = (
        ('Datastore/all', 12),
        ('Datastore/allOther', 12),
        ('Datastore/Postgres/all', 12),
        ('Datastore/Postgres/allOther', 12),
        ('Datastore/statement/Postgres/%s/select' % DB_SETTINGS['table_name'], 1),
        ('Datastore/statement/Postgres/%s/insert' % DB_SETTINGS['table_name'], 1),
        ('Datastore/statement/Postgres/%s/update' % DB_SETTINGS['table_name'], 1),
        ('Datastore/statement/Postgres/%s/delete' % DB_SETTINGS['table_name'], 1),
        ('Datastore/operation/Postgres/select', 1),
        ('Datastore/operation/Postgres/insert', 1),
        ('Datastore/operation/Postgres/update', 1),
        ('Datastore/operation/Postgres/delete', 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1),
        ('Datastore/statement/Postgres/now/call', 1),
        ('Datastore/statement/Postgres/pg_sleep/call', 1),
        ('Datastore/operation/Postgres/call', 2),
        ('Datastore/operation/Postgres/commit', 2),
        ('Datastore/operation/Postgres/rollback', 1),
)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(DB_SETTINGS['host'])
_port = DB_SETTINGS['port']

_instance_metric_name = 'Datastore/instance/Postgres/%s/%s' % (_host, _port)

_enable_rollup_metrics.append(
        (_instance_metric_name, 11)
)

_disable_rollup_metrics.append(
        (_instance_metric_name, None)
)


# Query
def _execute(connection, cursor, row_type, wrapper):
    unicode_type = psycopg2.extensions.UNICODE
    psycopg2.extensions.register_type(unicode_type)
    psycopg2.extensions.register_type(unicode_type, connection)
    psycopg2.extensions.register_type(unicode_type, cursor)

    sql = """drop table if exists %s""" % DB_SETTINGS["table_name"]
    cursor.execute(wrapper(sql))

    sql = """create table %s (a integer, b real, c text)""" % DB_SETTINGS["table_name"]
    cursor.execute(wrapper(sql))

    sql = """insert into %s """ % DB_SETTINGS["table_name"] + """values (%s, %s, %s)"""
    params = [(1, 1.0, '1.0'), (2, 2.2, '2.2'), (3, 3.3, '3.3')]
    cursor.executemany(wrapper(sql), params)

    sql = """select * from %s""" % DB_SETTINGS["table_name"]
    cursor.execute(wrapper(sql))

    for row in cursor:
        assert isinstance(row, row_type)

    sql = """update %s""" % DB_SETTINGS["table_name"] + """ set a=%s, b=%s, c=%s where a=%s"""
    params = (4, 4.0, '4.0', 1)
    cursor.execute(wrapper(sql), params)

    sql = """delete from %s where a=2""" % DB_SETTINGS["table_name"]
    cursor.execute(wrapper(sql))

    connection.commit()

    cursor.callproc('now')
    cursor.callproc('pg_sleep', (0.0,))

    connection.rollback()
    connection.commit()


def _exercise_db(cursor_factory=None, use_cur_context=False, row_type=tuple,
        wrapper=str):

    connection = psycopg2.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'])
    kwargs = {'cursor_factory': cursor_factory} if cursor_factory else {}

    try:
        if use_cur_context:
            with connection.cursor(**kwargs) as cursor:
                _execute(connection, cursor, row_type, wrapper)
        else:
            cursor = connection.cursor(**kwargs)
            _execute(connection, cursor, row_type, wrapper)
    finally:
        connection.close()


_test_matrix = ['wrapper,use_cur_context', [(str, False)]]


# with statement support for connections/cursors added in 2.5 and up
_test_matrix[1].append((str, True))

# Composable SQL is expected to be available in versions 2.7 and up
assert sql, (
        "Composable sql (from psycopg2 import sql) is expected to load"
        "but is not loading")

# exercise with regular SQL wrapper
_test_matrix[1].append((sql.SQL, True))
_test_matrix[1].append((sql.SQL, False))

# exercise with "Composed" SQL object
_test_matrix[1].append((lambda q: sql.Composed([sql.SQL(q)]), True))
_test_matrix[1].append((lambda q: sql.Composed([sql.SQL(q)]), False))


# Tests
@pytest.mark.parametrize(*_test_matrix)
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
        'test_cursor:test_execute_via_cursor_enable_instance',
        scoped_metrics=_enable_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor_enable_instance(wrapper, use_cur_context):
    _exercise_db(cursor_factory=None, use_cur_context=use_cur_context,
            row_type=tuple, wrapper=wrapper)


@pytest.mark.parametrize(*_test_matrix)
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
        'test_cursor:test_execute_via_cursor_disable_instance',
        scoped_metrics=_disable_scoped_metrics,
        rollup_metrics=_disable_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor_disable_instance(wrapper, use_cur_context):
    _exercise_db(cursor_factory=None, use_cur_context=use_cur_context,
            row_type=tuple, wrapper=wrapper)


@pytest.mark.parametrize(*_test_matrix)
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
        'test_cursor:test_execute_via_cursor_dict_enable_instance',
        scoped_metrics=_enable_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor_dict_enable_instance(wrapper, use_cur_context):
    dict_factory = psycopg2.extras.RealDictCursor
    _exercise_db(cursor_factory=dict_factory, use_cur_context=use_cur_context,
            row_type=dict, wrapper=wrapper)


@pytest.mark.parametrize(*_test_matrix)
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
        'test_cursor:test_execute_via_cursor_dict_disable_instance',
        scoped_metrics=_disable_scoped_metrics,
        rollup_metrics=_disable_rollup_metrics,
        background_task=True)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor_dict_disable_instance(wrapper, use_cur_context):
    dict_factory = psycopg2.extras.RealDictCursor
    _exercise_db(cursor_factory=dict_factory, use_cur_context=use_cur_context,
            row_type=dict, wrapper=wrapper)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

try:
    import urllib.request as urllib2
except:
    import urllib2

from testing_support.external_fixtures import (
    cache_outgoing_headers,
    insert_incoming_headers,
)
from testing_support.fixtures import cat_enabled
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_cross_process_headers import (
    validate_cross_process_headers,
)
from testing_support.validators.validate_external_node_params import (
    validate_external_node_params,
)

from newrelic.api.background_task import background_task


@pytest.fixture(scope="session")
def metrics(server):
    scoped = [("External/localhost:%d/urllib2/" % server.port, 1)]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("External/localhost:%d/urllib2/" % server.port, 1),
    ]

    return scoped, rollup


def test_urlopen_http_request(server, metrics):
    @validate_transaction_metrics(
        "test_urllib2:test_urlopen_http_request",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_urllib2:test_urlopen_http_request")
    def _test():
        urllib2.urlopen("http://localhost:%d/" % server.port)

    _test()


def test_urlopen_https_request(server, metrics):
    @validate_transaction_metrics(
        "test_urllib2:test_urlopen_https_request",
        scoped_metrics=metrics[0],
        rollup_metrics=metrics[1],
        background_task=True,
    )
    @background_task(name="test_urllib2:test_urlopen_https_request")
    def _test():
        try:
            urllib2.urlopen("https://localhost:%d/" % server.port)
        except Exception:
            pass

    _test()


def test_urlopen_http_request_with_port(server):
    scoped = [("External/localhost:%d/urllib2/" % server.port, 1)]

    rollup = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("External/localhost:%d/urllib2/" % server.port, 1),
    ]

    @validate_transaction_metrics(
        "test_urllib2:test_urlopen_http_request_with_port",
        scoped_metrics=scoped,
        rollup_metrics=rollup,
        background_task=True,
    )
    @background_task(name="test_urllib2:test_urlopen_http_request_with_port")
    def _test():
        urllib2.urlopen("http://localhost:%d/" % server.port)

    _test()


_test_urlopen_file_request_scoped_metrics = [("External/unknown/urllib2/", None)]

_test_urlopen_file_request_rollup_metrics = [
    ("External/all", None),
    ("External/allOther", None),
    ("External/unknown/urllib2/", None),
]


@validate_transaction_metrics(
    "test_urllib2:test_urlopen_file_request",
    scoped_metrics=_test_urlopen_file_request_scoped_metrics,
    rollup_metrics=_test_urlopen_file_request_rollup_metrics,
    background_task=True,
)
@background_task()
def test_urlopen_file_request():
    path = os.path.abspath(__file__)
    file_uri = "file://%s" % path
    urllib2.urlopen(file_uri)


@background_task()
@cache_outgoing_headers
@validate_cross_process_headers
def test_urlopen_cross_process_request(server):
    urllib2.urlopen("http://localhost:%d/" % server.port)


@cat_enabled
def test_urlopen_cross_process_response(server):
    _test_urlopen_cross_process_response_scoped_metrics = [
        ("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1)
    ]

    _test_urlopen_cross_process_response_rollup_metrics = [
        ("External/all", 1),
        ("External/allOther", 1),
        ("External/localhost:%d/all" % server.port, 1),
        ("ExternalApp/localhost:%d/1#2/all" % server.port, 1),
        ("ExternalTransaction/localhost:%d/1#2/test" % server.port, 1),
    ]

    _test_urlopen_cross_process_response_external_node_params = [
        ("cross_process_id", "1#2"),
        ("external_txn_name", "test"),
        ("transaction_guid", "0123456789012345"),
    ]

    @validate_transaction_metrics(
        "test_urllib2:test_urlopen_cross_process_response",
        scoped_metrics=_test_urlopen_cross_process_response_scoped_metrics,
        rollup_metrics=_test_urlopen_cross_process_response_rollup_metrics,
        background_task=True,
    )
    @insert_incoming_headers
    @validate_external_node_params(params=_test_urlopen_cross_process_response_external_node_params)
    @background_task(name="test_urllib2:test_urlopen_cross_process_response")
    def _test():
        urllib2.urlopen("http://localhost:%d/" % server.port)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pymysql
from testing_support.db_settings import mysql_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_database_trace_inputs import (
    validate_database_trace_inputs,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task

DB_SETTINGS = mysql_settings()[0]
TABLE_NAME = "datastore_pymysql_" + DB_SETTINGS["namespace"]
PROCEDURE_NAME = "hello_" + DB_SETTINGS["namespace"]

HOST = instance_hostname(DB_SETTINGS["host"])
PORT = DB_SETTINGS["port"]


def execute_db_calls_with_cursor(cursor):
    cursor.execute("""drop table if exists %s""" % TABLE_NAME)

    cursor.execute("""create table %s """ % TABLE_NAME + """(a integer, b real, c text)""")

    cursor.executemany(
        """insert into %s """ % TABLE_NAME + """values (%s, %s, %s)""",
        [(1, 1.0, "1.0"), (2, 2.2, "2.2"), (3, 3.3, "3.3")],
    )

    cursor.execute("""select * from %s""" % TABLE_NAME)

    for row in cursor:
        pass

    cursor.execute("""update %s""" % TABLE_NAME + """ set a=%s, b=%s, """ """c=%s where a=%s""", (4, 4.0, "4.0", 1))

    cursor.execute("""delete from %s where a=2""" % TABLE_NAME)
    cursor.execute("""drop procedure if exists %s""" % PROCEDURE_NAME)
    cursor.execute(
        """CREATE PROCEDURE %s()
                      BEGIN
                        SELECT 'Hello World!';
                      END"""
        % PROCEDURE_NAME
    )

    cursor.callproc(PROCEDURE_NAME)


_test_execute_via_cursor_scoped_metrics = [
    ("Function/pymysql:Connect", 1),
    ("Datastore/statement/MySQL/%s/select" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/insert" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/update" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/delete" % TABLE_NAME, 1),
    ("Datastore/operation/MySQL/drop", 2),
    ("Datastore/operation/MySQL/create", 2),
    ("Datastore/statement/MySQL/%s/call" % PROCEDURE_NAME, 1),
    ("Datastore/operation/MySQL/commit", 2),
    ("Datastore/operation/MySQL/rollback", 1),
]

_test_execute_via_cursor_rollup_metrics = [
    ("Datastore/all", 13),
    ("Datastore/allOther", 13),
    ("Datastore/MySQL/all", 13),
    ("Datastore/MySQL/allOther", 13),
    ("Datastore/statement/MySQL/%s/select" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/insert" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/update" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/delete" % TABLE_NAME, 1),
    ("Datastore/operation/MySQL/select", 1),
    ("Datastore/operation/MySQL/insert", 1),
    ("Datastore/operation/MySQL/update", 1),
    ("Datastore/operation/MySQL/delete", 1),
    ("Datastore/statement/MySQL/%s/call" % PROCEDURE_NAME, 1),
    ("Datastore/operation/MySQL/call", 1),
    ("Datastore/operation/MySQL/drop", 2),
    ("Datastore/operation/MySQL/create", 2),
    ("Datastore/operation/MySQL/commit", 2),
    ("Datastore/operation/MySQL/rollback", 1),
    ("Datastore/instance/MySQL/%s/%s" % (HOST, PORT), 12),
]


@validate_transaction_metrics(
    "test_database:test_execute_via_cursor",
    scoped_metrics=_test_execute_via_cursor_scoped_metrics,
    rollup_metrics=_test_execute_via_cursor_rollup_metrics,
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor():
    connection = pymysql.connect(
        db=DB_SETTINGS["name"],
        user=DB_SETTINGS["user"],
        passwd=DB_SETTINGS["password"],
        host=DB_SETTINGS["host"],
        port=DB_SETTINGS["port"],
    )

    with connection.cursor() as cursor:
        execute_db_calls_with_cursor(cursor)

    connection.commit()
    connection.rollback()
    connection.commit()


_test_execute_via_cursor_context_mangaer_scoped_metrics = [
    ("Function/pymysql:Connect", 1),
    ("Datastore/statement/MySQL/%s/select" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/insert" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/update" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/delete" % TABLE_NAME, 1),
    ("Datastore/operation/MySQL/drop", 2),
    ("Datastore/operation/MySQL/create", 2),
    ("Datastore/statement/MySQL/%s/call" % PROCEDURE_NAME, 1),
    ("Datastore/operation/MySQL/commit", 2),
    ("Datastore/operation/MySQL/rollback", 1),
]

_test_execute_via_cursor_context_mangaer_rollup_metrics = [
    ("Datastore/all", 13),
    ("Datastore/allOther", 13),
    ("Datastore/MySQL/all", 13),
    ("Datastore/MySQL/allOther", 13),
    ("Datastore/statement/MySQL/%s/select" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/insert" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/update" % TABLE_NAME, 1),
    ("Datastore/statement/MySQL/%s/delete" % TABLE_NAME, 1),
    ("Datastore/operation/MySQL/select", 1),
    ("Datastore/operation/MySQL/insert", 1),
    ("Datastore/operation/MySQL/update", 1),
    ("Datastore/operation/MySQL/delete", 1),
    ("Datastore/statement/MySQL/%s/call" % PROCEDURE_NAME, 1),
    ("Datastore/operation/MySQL/call", 1),
    ("Datastore/operation/MySQL/drop", 2),
    ("Datastore/operation/MySQL/create", 2),
    ("Datastore/operation/MySQL/commit", 2),
    ("Datastore/operation/MySQL/rollback", 1),
    ("Datastore/instance/MySQL/%s/%s" % (HOST, PORT), 12),
]


@validate_transaction_metrics(
    "test_database:test_execute_via_cursor_context_manager",
    scoped_metrics=_test_execute_via_cursor_context_mangaer_scoped_metrics,
    rollup_metrics=_test_execute_via_cursor_context_mangaer_rollup_metrics,
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor_context_manager():
    connection = pymysql.connect(
        db=DB_SETTINGS["name"],
        user=DB_SETTINGS["user"],
        passwd=DB_SETTINGS["password"],
        host=DB_SETTINGS["host"],
        port=DB_SETTINGS["port"],
    )
    cursor = connection.cursor()

    with cursor:
        execute_db_calls_with_cursor(cursor)

    connection.commit()
    connection.rollback()
    connection.commit()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

import pytest
import webtest
from testing_support.fixtures import override_application_settings, validate_attributes
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.transaction import (
    accept_distributed_trace_headers,
    current_transaction,
    insert_distributed_trace_headers,
)
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.encoding_utils import W3CTraceState
from newrelic.common.object_wrapper import transient_function_wrapper
from newrelic.packages import six

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
JSON_DIR = os.path.normpath(os.path.join(CURRENT_DIR, "fixtures", "distributed_tracing"))

_parameters_list = (
    "test_name",
    "trusted_account_key",
    "account_id",
    "web_transaction",
    "raises_exception",
    "force_sampled_true",
    "span_events_enabled",
    "transport_type",
    "inbound_headers",
    "outbound_payloads",
    "intrinsics",
    "expected_metrics",
)

_parameters = ",".join(_parameters_list)


XFAIL_TESTS = [
    "spans_disabled_root",
    "missing_traceparent",
    "missing_traceparent_and_tracestate",
    "w3c_and_newrelc_headers_present_error_parsing_traceparent",
]


def load_tests():
    result = []
    path = os.path.join(JSON_DIR, "trace_context.json")
    with open(path, "r") as fh:
        tests = json.load(fh)

    for test in tests:
        values = (test.get(param, None) for param in _parameters_list)
        param = pytest.param(*values, id=test.get("test_name"))
        result.append(param)

    return result


ATTR_MAP = {
    "traceparent.version": 0,
    "traceparent.trace_id": 1,
    "traceparent.parent_id": 2,
    "traceparent.trace_flags": 3,
    "tracestate.version": 0,
    "tracestate.parent_type": 1,
    "tracestate.parent_account_id": 2,
    "tracestate.parent_application_id": 3,
    "tracestate.span_id": 4,
    "tracestate.transaction_id": 5,
    "tracestate.sampled": 6,
    "tracestate.priority": 7,
    "tracestate.timestamp": 8,
    "tracestate.tenant_id": None,
}


def validate_outbound_payload(actual, expected, trusted_account_key):
    traceparent = ""
    tracestate = ""
    for key, value in actual:
        if key == "traceparent":
            traceparent = value.split("-")
        elif key == "tracestate":
            vendors = W3CTraceState.decode(value)
            nr_entry = vendors.pop(trusted_account_key + "@nr", "")
            tracestate = nr_entry.split("-")
    exact_values = expected.get("exact", {})
    expected_attrs = expected.get("expected", [])
    unexpected_attrs = expected.get("unexpected", [])
    expected_vendors = expected.get("vendors", [])
    for key, value in exact_values.items():
        header = traceparent if key.startswith("traceparent.") else tracestate
        attr = ATTR_MAP[key]
        if attr is not None:
            if isinstance(value, bool):
                assert header[attr] == str(int(value))
            elif isinstance(value, int):
                assert int(header[attr]) == value
            else:
                assert header[attr] == str(value)

    for key in expected_attrs:
        header = traceparent if key.startswith("traceparent.") else tracestate
        attr = ATTR_MAP[key]
        if attr is not None:
            assert header[attr], key

    for key in unexpected_attrs:
        header = traceparent if key.startswith("traceparent.") else tracestate
        attr = ATTR_MAP[key]
        if attr is not None:
            assert not header[attr], key

    for vendor in expected_vendors:
        assert vendor in vendors


@wsgi_application()
def target_wsgi_application(environ, start_response):
    transaction = current_transaction()

    if not environ[".web_transaction"]:
        transaction.background_task = True

    if environ[".raises_exception"]:
        try:
            raise ValueError("oops")
        except:
            transaction.notice_error()

    if ".inbound_headers" in environ:
        accept_distributed_trace_headers(
            environ[".inbound_headers"],
            transport_type=environ[".transport_type"],
        )

    payloads = []
    for _ in range(environ[".outbound_calls"]):
        payloads.append([])
        insert_distributed_trace_headers(payloads[-1])

    start_response("200 OK", [("Content-Type", "application/json")])
    return [json.dumps(payloads).encode("utf-8")]


test_application = webtest.TestApp(target_wsgi_application)


def override_compute_sampled(override):
    @transient_function_wrapper("newrelic.core.adaptive_sampler", "AdaptiveSampler.compute_sampled")
    def _override_compute_sampled(wrapped, instance, args, kwargs):
        if override:
            return True
        return wrapped(*args, **kwargs)

    return _override_compute_sampled


@pytest.mark.parametrize(_parameters, load_tests())
def test_trace_context(
    test_name,
    trusted_account_key,
    account_id,
    web_transaction,
    raises_exception,
    force_sampled_true,
    span_events_enabled,
    transport_type,
    inbound_headers,
    outbound_payloads,
    intrinsics,
    expected_metrics,
):
    if test_name in XFAIL_TESTS:
        pytest.xfail("Waiting on cross agent tests update.")
    # Prepare assertions
    if not intrinsics:
        intrinsics = {}

    common = intrinsics.get("common", {})
    common_required = common.get("expected", [])
    common_forgone = common.get("unexpected", [])
    common_exact = common.get("exact", {})

    txn_intrinsics = intrinsics.get("Transaction", {})
    txn_event_required = {"agent": [], "user": [], "intrinsic": txn_intrinsics.get("expected", [])}
    txn_event_required["intrinsic"].extend(common_required)
    txn_event_forgone = {"agent": [], "user": [], "intrinsic": txn_intrinsics.get("unexpected", [])}
    txn_event_forgone["intrinsic"].extend(common_forgone)
    txn_event_exact = {"agent": {}, "user": {}, "intrinsic": txn_intrinsics.get("exact", {})}
    txn_event_exact["intrinsic"].update(common_exact)

    override_settings = {
        "distributed_tracing.enabled": True,
        "span_events.enabled": span_events_enabled,
        "account_id": account_id,
        "trusted_account_key": trusted_account_key,
    }

    extra_environ = {
        ".web_transaction": web_transaction,
        ".raises_exception": raises_exception,
        ".transport_type": transport_type,
        ".outbound_calls": outbound_payloads and len(outbound_payloads) or 0,
    }

    inbound_headers = inbound_headers and inbound_headers[0] or None
    if transport_type != "HTTP":
        extra_environ[".inbound_headers"] = inbound_headers
        inbound_headers = None
    elif six.PY2 and inbound_headers:
        inbound_headers = {k.encode("utf-8"): v.encode("utf-8") for k, v in inbound_headers.items()}

    @validate_transaction_metrics(
        test_name, group="Uri", rollup_metrics=expected_metrics, background_task=not web_transaction
    )
    @validate_transaction_event_attributes(txn_event_required, txn_event_forgone, txn_event_exact)
    @validate_attributes("intrinsic", common_required, common_forgone)
    @override_application_settings(override_settings)
    @override_compute_sampled(force_sampled_true)
    def _test():
        return test_application.get(
            "/" + test_name,
            headers=inbound_headers,
            extra_environ=extra_environ,
        )

    if "Span" in intrinsics:
        span_intrinsics = intrinsics.get("Span")
        span_expected = span_intrinsics.get("expected", [])
        span_expected.extend(common_required)
        span_unexpected = span_intrinsics.get("unexpected", [])
        span_unexpected.extend(common_forgone)
        span_exact = span_intrinsics.get("exact", {})
        span_exact.update(common_exact)

        _test = validate_span_events(
            exact_intrinsics=span_exact, expected_intrinsics=span_expected, unexpected_intrinsics=span_unexpected
        )(_test)
    elif not span_events_enabled:
        _test = validate_span_events(count=0)(_test)

    response = _test()
    assert response.status == "200 OK"
    payloads = response.json
    if outbound_payloads:
        assert len(payloads) == len(outbound_payloads)
        for actual, expected in zip(payloads, outbound_payloads):
            validate_outbound_payload(actual, expected, trusted_account_key)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from ariadne import (
    MutationType,
    QueryType,
    UnionType,
    load_schema_from_path,
    make_executable_schema,
)
from ariadne.asgi import GraphQL as GraphQLASGI
from framework_graphql._target_schema_sync import books, magazines, libraries

from testing_support.asgi_testing import AsgiTest

schema_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "schema.graphql")
type_defs = load_schema_from_path(schema_file)

storage = []

mutation = MutationType()


@mutation.field("storage_add")
async def resolve_storage_add(self, info, string):
    storage.append(string)
    return string


item = UnionType("Item")


@item.type_resolver
async def resolve_type(obj, *args):
    if "isbn" in obj:
        return "Book"
    elif "issue" in obj:  # pylint: disable=R1705
        return "Magazine"

    return None


query = QueryType()


@query.field("library")
async def resolve_library(self, info, index):
    return libraries[index]


@query.field("storage")
async def resolve_storage(self, info):
    return [storage.pop()]


@query.field("search")
async def resolve_search(self, info, contains):
    search_books = [b for b in books if contains in b["name"]]
    search_magazines = [m for m in magazines if contains in m["name"]]
    return search_books + search_magazines


@query.field("hello")
@query.field("error_middleware")
async def resolve_hello(self, info):
    return "Hello!"


@query.field("echo")
async def resolve_echo(self, info, echo):
    return echo


@query.field("error_non_null")
@query.field("error")
async def resolve_error(self, info):
    raise RuntimeError("Runtime Error!")


target_schema = make_executable_schema(type_defs, query, mutation, item)
target_asgi_application = AsgiTest(GraphQLASGI(target_schema))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2
import pytest

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_slow_sql_collector_json import validate_slow_sql_collector_json

from utils import DB_SETTINGS

from newrelic.api.background_task import background_task
from newrelic.api.transaction import current_transaction


# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
    'datastore_tracer.database_name_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
    'datastore_tracer.database_name_reporting.enabled': False,
}

# Expected parameters

_enabled_required = set(['host', 'port_path_or_id', 'database_name'])
_enabled_forgone = set()

_disabled_required = set()
_disabled_forgone = set(['host', 'port_path_or_id', 'database_name'])

_distributed_tracing_always_params = set(['guid', 'traceId', 'priority',
    'sampled'])
_distributed_tracing_payload_received_params = set(['parent.type',
    'parent.app', 'parent.account', 'parent.transportType',
    'parent.transportDuration'])

_transaction_guid = '1234567890'
_distributed_tracing_exact_params = {'guid': _transaction_guid}


# Query

def _exercise_db():
    connection = psycopg2.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'])

    try:
        cursor = connection.cursor()
        cursor.execute("""SELECT setting from pg_settings where name=%s""",
                ('server_version',))
    finally:
        connection.close()


# Tests

@pytest.mark.parametrize('instance_enabled', (True, False))
@pytest.mark.parametrize('distributed_tracing_enabled,payload_received', [
        (True, True),
        (True, False),
        (False, False),
])
def test_slow_sql_json(instance_enabled, distributed_tracing_enabled,
        payload_received):

    exact_params = None

    if instance_enabled:
        settings = _enable_instance_settings.copy()
        required_params = set(_enabled_required)
        forgone_params = set(_enabled_forgone)
    else:
        settings = _disable_instance_settings.copy()
        required_params = set(_disabled_required)
        forgone_params = set(_disabled_forgone)

    if distributed_tracing_enabled:
        required_params.update(_distributed_tracing_always_params)
        exact_params = _distributed_tracing_exact_params
        settings['distributed_tracing.enabled'] = True
        if payload_received:
            required_params.update(
                    _distributed_tracing_payload_received_params)
        else:
            forgone_params.update(
                    _distributed_tracing_payload_received_params)
    else:
        forgone_params.update(_distributed_tracing_always_params)
        forgone_params.update(_distributed_tracing_payload_received_params)
        settings['distributed_tracing.enabled'] = False

    @override_application_settings(settings)
    @validate_slow_sql_collector_json(
            required_params=required_params,
            forgone_params=forgone_params,
            exact_params=exact_params)
    @background_task()
    def _test():
        transaction = current_transaction()
        transaction.guid = _transaction_guid

        _exercise_db()

        if payload_received:

            payload = {
                'v': [0, 1],
                'd': {
                    'ty': 'Mobile',
                    'ac': transaction.settings.account_id,
                    'tk': transaction.settings.trusted_account_key,
                    'ap': '2827902',
                    'pa': '5e5733a911cfbc73',
                    'id': '7d3efb1b173fecfa',
                    'tr': 'd6b4ba0c3a712ca',
                    'ti': 1518469636035,
                    'tx': '8703ff3d88eefe9d',
                }
            }

            transaction.accept_distributed_trace_payload(payload)

    _test()
